package rt ; 

require Exporter;   
@ISA = ('Exporter');
@EXPORT=qw(rtchildren rtalarm rtbkg rtkill rtrun rtcmdout mbkg mwait showmbkg rtcompress) ;

use strict ;
use warnings ; 
use diagnostics ;

use lib "$ENV{ANTELOPE}/data/perl" ;

use children ;
use staticsys ;

use POSIX ":sys_wait_h";
use POSIX qw(setsid setpgid) ;
use IO::Handle ;

our $Concurrent = -1 ;

sub rtrun {
    my ( $cmd, $timeout, $stdout, $stderr ) = @_ ;
    my $pid ; 
    my $aborted = 0 ;
    if (defined $timeout && $timeout > 0) {        
	$SIG{ALRM} = sub { 
	    print STDERR "\n\n  **rtrun kills pid $pid ($cmd) after timeout of $timeout seconds\n" ;
	    rtkill (1, $pid) ;
	    $aborted = 1 ;
	};
	alarm $timeout ;
    }
    $pid = fork ;
    if ( $pid ) {
	# parent
    } elsif ( defined $pid ) {
	# child
	setsid() ;
	open ( STDIN, "</dev/null" ) ;
	if (defined $stdout) { 
	    open ( STDOUT, ">$stdout" ) ;
	    STDOUT->autoflush(1);
	    if (defined $stderr) { 
		open ( STDERR, ">$stderr" ) ;
		STDERR->autoflush(1);
	    } else { 
		open ( STDERR, ">&STDOUT" ) ;
	    }
	}
	exec $cmd 
	    or print STDERR "Can't exec '$cmd' : $!\n" ; 
	exit 0 ; 
    } else {
	# fork failure
        print STDERR "Can't fork '$cmd'\n" ; 
	exit 0 ; 
    }
    waitpid ( $pid, 0 ) ; 
    my $left  = alarm 0 if defined $timeout ;
    return $aborted ;
}

sub rtbkg { 
    my ( $cmd, $stdout, $stderr ) = @_ ; 
    my ( $pid ) ;
    if ( $pid = fork ) { 
	# parent
	return  $pid ; 
    } elsif ( defined $pid ) {
	# child
	setsid() ;
	open ( STDIN, "/dev/null" ) ; 
	if (defined $stdout) { 
	    open ( STDOUT, ">$stdout" ) or
	    	warn "Can't open '$stdout' for writing : $!\n" ; 
	    STDOUT->autoflush(1);
	    if (defined $stderr) { 
		open ( STDERR, ">$stderr" ) or
		    warn "Can't open '$stderr' for writing : $!\n" ; 
		STDERR->autoflush(1);
	    } else { 
		open ( STDERR, ">&STDOUT" ) ;
	    }
	}
	# print STDERR "exec'ing '$cmd'\n" ;
        exec $cmd 
	    or print STDERR ( "Can't exec '$cmd' : $!\n" ) ; 
	exit 1 ; 
    } else {
	# fork failure
        print STDERR "Can't fork '$cmd'\n" ; 
	return -1 ; 
    }
}

sub rtalarm { 
    my ( $pid, $timeout, $ref ) = @_ ;
    our $RtAlarm = $ref ; 
    our $RtPid = $pid ; 
    our $RtTimeout = $timeout ;
    $$RtAlarm = 0 ; 

    $SIG{'ALRM'} = sub { 
	print STDERR "Timeout after $RtTimeout seconds: Killing $RtPid\n" ; 
	$$RtAlarm = 1 ; 
        rtkill (1, $pid) ;
    };

    alarm $timeout ;       
}

sub rtkill {
    my ($time2die, @pids) = @_ ;
    my ($pid) ;
    my (@all) = (); 
    my (@all2) = () ;
    my (@subs, $i, @children) ;
    my (@msgs) = () ; 
    return if (@pids == 0 ) ;
    foreach $pid ( @pids ) {
#  suppose rtkill is called with undef: "rtkill (undef);"
	if ( defined $pid && pid_exists($pid) ) {
	    @children = &rtchildren($pid) ;
	    if ( @children > 0 ) {
		push(@msgs, "killing pid $pid and its children @children") ;
	    } else { 
		push(@msgs, "killing pid $pid") ;
	    }
	    push ( @all, $pid, @children ) ;
	    kill 'SIGTERM', $pid, @children ;
	}
    }

    for ( $i=0 ; @all > 0 && $i<$time2die ; $i++ ) {
	sleep 1 ;
	foreach $pid ( @all ) {
	    waitpid ( $pid, &WNOHANG ) ; 
	}
	@all2 = () ;
	foreach $pid ( @all ) {
	    if ( pid_exists($pid) ) {
		push (@all2, $pid ) ; 
	    }
	}
	@all = @all2 ; 
	if ( $i > 1 && $i % 1 == 0 ) { 
	    if ( @children > 0 ) {
		push(@msgs, "waiting for pids @all" ) ; 
	    } else { 
		push(@msgs, "waiting for pids @all (pid and its children)" ) ; 
	    }
	}
    }

    if ( @all > 0 ) { 
	foreach $pid ( @all ) {
	    push(@msgs, "forced to kill -9 $pid") ;
	    kill '9', $pid ;
	}
	sleep 1 ;
	foreach $pid ( @all ) {
	    waitpid ( $pid, &WNOHANG ) ; 
	}

	foreach $pid ( @all ) {
	    if ( pid_exists($pid) ) {
		push(@msgs, "can't kill $pid" ) ;
	    }
	}
    }
    return @msgs ;
}

sub rtcmdout {
    my ( $cmd, $timeout ) = @_ ;
    my $pid ; 
    my $aborted = 0 ;

    if ( defined $timeout ) { 
	local $SIG{'ALRM'} = sub { 
	    rtkill (1, $pid) ;
	    $aborted = 1 ;
	};

	alarm $timeout ;       
    }

    my @args = split(' ', $cmd) ;
    my $output = "" ;
    my $line ; 
    if ($pid = open(CMD, "-|")) {
        while ($line=<CMD>) {
            $output .= $line;
        }
        close CMD;
    } else {
        open (STDERR, ">&STDOUT") ;
        exec { $args[0] } @args 
            or print STDERR "$!\n" ;  
        exit 1 ;
    }
    my $result = $? ;
    waitpid ( $pid, 0 ) ; 
    my $left  = alarm 0 if defined $timeout ;
    return ($result, $output) ;
}

sub by_number { 
    return $a <=> $b ; 
}

sub showmbkg { 
    my ($hashref) = @_ ;
    print STDERR "\nshow background queue info: $Concurrent tasks possible\n" ; 
    my $run = $$hashref{RUN} ; 
    print STDERR "  run is $run\n" ;
    my $logdir = $$hashref{LOGS} ;
    print STDERR "  logdir = $logdir\n" ; 
    my $pidsref = $$hashref{PIDS} ; 
    printf STDERR "  %d Tasks Running:\n", scalar(keys %$pidsref) ;
    foreach my $pid ( sort by_number keys %$pidsref ) { 
	printf STDERR "    %5d %s\n", $pid, $$pidsref{$pid} ; 
    }
    my $queueref = $$hashref{QUEUE} ; 
    if ( @$queueref > 0 ) { 
	print STDERR "  Waiting to run :\n" ; 
	foreach my $cmd (@$queueref) { 
	    print STDERR "    --> $cmd\n" ; 
	}
    } else { 
	print STDERR "  Wait queue is empty\n" ; 
    }

    print STDERR "  All background tasks\n" ;
    my $cmdref = $$hashref{CMD} ;
    foreach my $id ( sort by_number keys %$cmdref ) { 
	printf STDERR "    %3d %s\n", $id, $$cmdref{$id} ; 
    }
    print STDERR "\n\n" ;
}

sub mbkg { 
    my ( $hashref, $cmd ) = @_ ; 
    if ( ! defined $$hashref{RUN} ) { 
	$$hashref{QUEUE} = [] ;
	$$hashref{CMD} = {} ; 
	$$hashref{PIDS} = {} ; 
	my $logdir = $$hashref{LOGS} = "/tmp/mbkg$$" ; 
	if ( ! -d $logdir ) { 
	    mkdir $logdir, 0775 or
		    die "Can't create $logdir : $!\n" ; 
	}
	if ( $Concurrent < 1 ) { 
	    $Concurrent = 1 ;
	    chomp($Concurrent = `getid concurrent` ) ;
	}
	$$hashref{FIRST} = $$hashref{RUN} = 0 ;
    }
    my $pidsref = $$hashref{PIDS} ; 
    my $cmdref = $$hashref{CMD} ;
    for (;;) { 
	my $pid = waitpid(-1, WNOHANG) ; 
	last if $pid <= 0 ;
	delete $$pidsref{$pid} ; 
    }
    my $running = scalar(keys %$pidsref) ; 
    my $queueref = $$hashref{QUEUE} ; 
    my $logdir = $$hashref{LOGS} ;
    push (@$queueref, $cmd) if defined $cmd ;
    while ( @$queueref > 0 && $running < $Concurrent) { 
	my $cmd = shift(@$queueref) ; 
	my $run = $$hashref{RUN} ; 
	my $stdout = "$logdir/$run" ;
	while ( -e $stdout ) { 
	    $run++ ;
	    $$hashref{RUN} = $run ;
	    $stdout = "$logdir/$$hashref{RUN}" ;
	}
	my $pid = rtbkg($cmd, $stdout) ;
	if ( $pid > 0 ) { 
	    $running++ ;
	    $$pidsref{$pid} = $cmd ;
	    $$cmdref{$run} = $cmd ;
	    $$hashref{RUN} = ++$run ;
	} else { 
	    print STDERR "failed to run '$cmd'\n" ;
	}
    }
}

sub mwait { 
    my ($hashref, $keep) = @_ ; 
    return if ! defined $$hashref{PIDS} ; 
    my $pidsref = $$hashref{PIDS} ; 
    my $queueref = $$hashref{QUEUE} ; 
    # my $cnt = 0 ; 
    for (;;) { 
	my $pid = waitpid(-1, 0) ; 
	mbkg($hashref, undef) if @$queueref > 0 ;  # force queue to run
	delete $$pidsref{$pid} ; 
	# printf STDERR "iteration #%d\n", $cnt++ ; 
	# showmbkg ($hashref) ;
	last if scalar(keys %$pidsref) < 1 && @$queueref < 1 ; 
    }
    # printf STDERR "Quit: running = %d  waiting = %d\n", scalar(keys %$pidsref), scalar(@$queueref) ; 
    my $logdir = $$hashref{LOGS} ;
    my $cmdref = $$hashref{CMD} ;
    my @whole ;
    for ( my $run= 0  ; $run<$$hashref{RUN} ; $run++ ) { 
	if ( defined $$cmdref{$run} ) { 
	    my $log = "$logdir/$run" ; 
	    open ( LOG, $log ) 
	    	or warn "Can't open $log : $@\n" ;
	    my @log = <LOG> ; 
	    push (@whole, "\n--> $$cmdref{$run}\n", @log) ; 
	} else { 
	    print STDERR "No record of run #$run !!!\n" ; 
	}
    }
    if ( ! defined $keep ) { 
	system ( "rm -rf $logdir" ) ; 
    }
    return @whole ; 
}


sub rtcompress { 
    my ($filename) = @_ ; 
    if ( -e $filename ) { 
	system ( "bzip2 $filename" ) ; 
	return "$filename.bz2" ; 
    }
}

1;
